#include <Python.h>
#include <casa/aips.h>
#include <casa/Arrays.h>
#include <casa/Containers/Record.h>
#include <casa/Arrays/ArrayPosIter.h>
#include <tables/Tables.h>
#include <tables/Tables/BaseTable.h>
#include <tables/Tables/TableParse.h>
#include <tables/Tables/TableKeyword.h>

#include "stringifier.h"

#include <boost/python.hpp>

#include <arrayobject.h>

#include <algorithm>
#include <map>


using namespace casa;
using namespace std;
using namespace boost::python;

// Table cruft
struct column_proxy_no_such_column:
    std::exception 
{};
struct column_proxy_unsupported_column_datatype:
    std::exception 
{};

struct failed_to_load_Numarray_API:
    std::exception
{};

struct failed_to_create_Numarray_object:
    std::exception
{};

// mapping of casa::datatype -> python Numeric datatype
typedef std::map<casa::DataType,PyArray_TYPES>   dtmap_t;

dtmap_t make_dtmap( void )
{
    dtmap_t  rv;

    rv[TpBool]       = PyArray_INT;
    rv[TpInt]        = PyArray_INT;
    rv[TpUInt]       = PyArray_UINT;
    rv[TpString]     = PyArray_OBJECT;
    rv[TpDouble]     = PyArray_DOUBLE;
    rv[TpFloat]      = PyArray_FLOAT;
    rv[TpComplex]    = PyArray_CFLOAT;
    rv[TpDComplex]   = PyArray_CDOUBLE;
    rv[TpArrayFloat] = PyArray_OBJECT;
    return rv;
}

dtmap_t   dtmap = make_dtmap();

namespace warp {
    template <typename T>
    struct identity {
        typedef T  returntype;
        T operator()( const T& t ) const {
            return t;
        }

        // warp a number of objects in one go
        // Since this is the *identity* warper
        // it assumes T is POD [plain old data]
        // hence that bitwise copies are Ok
        // For objects which must use 'operator='
        // to work properly, we must implement
        // a different warper. For now it's not
        // necessary so...
        void operator()(T* dst, const T* src, unsigned int n) const {
            ::memcpy(dst, src, n*sizeof(T));
            return;
        }
    };

    // the template parameter here is unused but at least
    // now this thang can be used as template template
    // parameter
    template <typename T=int>
    struct ToPyStr {
        typedef PyObject* returntype;
        PyObject* operator()( const casa::String& t ) const {
            return PyString_FromString(t.chars());
        }

        // nothing else to do but to loop over all items...
        void operator()(returntype* dst, const casa::String* src, unsigned int n) const {
            for( unsigned int i=0; i<n; ++i ) {
                *dst++ = (*this)(*src++);
            }
            return;
        }
    };

    // for casa::Complex and casa::DComplex

    // this warper ignores the template argument
    // and converts whatever is passed to it
    // to an 'array of 2 floats'
    template <typename T>
    struct ToPyComplex {
        typedef float basic_type;
        typedef struct {
            basic_type   re;
            basic_type   im;
        } returntype;

        // Support anything that has .real() and .imag()
        // which are convertible to 'float'
        template <typename U>
        returntype operator()( const U& c ) const {
            returntype    rv;

            rv.re = basic_type(c.real());
            rv.im = basic_type(c.imag());
            return rv;
        }

        // first, the generic, slow implementation
        // which supports converting an array of
        // U's which have .real() and .imag()
        // which are convertible to float.
        template <typename U>
        void operator()(returntype* dst, const U* src, unsigned int n) const {
            for( unsigned int i=0; i<n; ++i ) {
                *dst++ = (*this)(*src++);
            }
            return;
        }

        // specialization for casa::Complex -> these are
        // in layout identical to PyComplex so we can
        // optimize now and just do a plain memcpy!
        void operator()(returntype* dst, const casa::Complex* src, unsigned int n) const {
            ::memcpy(dst, src, n*sizeof(returntype));
            return;
        }
    };

    template <typename T>
    struct ToPyDComplex {
        typedef double basic_type;
        typedef struct {
            basic_type   re;
            basic_type   im;
        } returntype;

        // Support anything that has .real() and .imag()
        // which are convertible to 'double'
        template <typename U>
        returntype operator()( const U& c ) const {
            returntype    rv;

            rv.re = basic_type(c.real());
            rv.im = basic_type(c.imag());
            return rv;
        }

        // first, the generic, slow implementation
        // which supports converting an array of
        // U's which have .real() and .imag()
        // which are convertible to double.
        template <typename U>
        void operator()(returntype* dst, const U* src, unsigned int n) const {
            for( unsigned int i=0; i<n; ++i ) {
                *dst++ = (*this)(*src++);
            }
            return;
        }

        // specialization for casa::DComplex -> these are
        // in layout identical to PyDComplex so we can
        // optimize now and just do a plain memcpy!
        void operator()(returntype* dst, const casa::DComplex* src, unsigned int n) const {
            ::memcpy(dst, src, n*sizeof(returntype));
            return;
        }
    };
}



//
// Transform casa::Array to a numarray
// 
template <template <typename> class Transform=warp::identity>
struct arraytransformer {
    template <typename T>
    PyObject* operator()( const Array<T>& t ) const
    {
        IPosition                           shp( t.shape() );
        IPosition                           newshp( shp.nelements() );
        Transform<T>                        warper;
        PyArrayObject*                      po;
        dtmap_t::const_iterator             dstdt;
        typename Transform<T>::returntype*  dst;

        // if we don't recognize the type, tell 'm so!
        // The datatypemap tells us, given the casa datatype,
        // what the corresponding Numarray datatype would be
        if( (dstdt=dtmap.find(whatType((T*)0)))==dtmap.end() )
        {
            // if this occurs, maybe you need to 
            // add an entry to the 'dtmap' variable....
            // browse this file and look for it.
            ostringstream s;
            s << __FILE__ << ":" << __LINE__ << " - arraytransformer<> -> Unhandled datatype "
                 << whatType((T*)0);
            cerr << s.str() << endl;
            PyErr_SetString(PyExc_ValueError,s.str().c_str());
            return NULL;
        }

        // ok, so we recognized the type, only the array's empty...
        if( shp.product()==0 ) {
            Py_INCREF(Py_None);
            return Py_None;
        }

        // if the PyArray API not loaded, then load it!
        if( !PyArray_API ) {
            import_array();
        }
        if( PyArray_API==0 ) {
            throw failed_to_load_Numarray_API();
        }

        // create the array from the dimensions
        po = (PyArrayObject*)PyArray_FromDims(shp.nelements(),
                                              (int*)shp.storage(),
                                              dstdt->second);

        if( po==0 ) {
            PyErr_SetString(PyExc_MemoryError, "Failed to allocate Numarray-object");
            return NULL;
        }

        // do a cast to destination type
        dst = (typename Transform<T>::returntype*)po->data;

        // do it like this to be able to *transform*
        // between different types...
        // Eg when transforming an array of strings;
        // each string has to be transformed to
        // a PyObject in that case...
        //
        // So we first get a contiguous array
        // out of the source array such that
        // we can loop as fast as possible
        // over the elements.
        //
        // The 'warper' will be called upon
        // to do the transformation.
        // There may be specializations
        // which do a memcpy, others use 
        // a for-loop with operator=()
        // [sometimes you just have to do it like
        //  that...]
        const T*   src;
        casa::Bool del;

        src = t.getStorage(del);
        warper(dst, src, shp.product());
        t.freeStorage(src, del);

        // copy the strides to fix
        // array-indexing
        // numarray strides are in units of bytes,
        // casa steps are in units of 'items'
        // so must multiply the casa strides by
        // the itemsize
        for( unsigned int k=0; k<t.steps().nelements(); ++k )
            po->strides[k] = t.steps()[k] * sizeof(typename Transform<T>::returntype);

        return (PyObject*)po;
    }
};


// convert all the arrayvalues
struct tostdstring
{
    std::string operator()( const casa::String& cs ) const 
    {
        return std::string(cs.chars());
    }
};


// Record-field accessor, gets the (scalar) valued field
// into the variable 'value'
template <typename T>
struct rfaccessor_t {
    rfaccessor_t(const casa::RecordInterface& r, const RecordFieldId& fi) {
        r.get(fi, value);
    }

    T   value;
};

// id. as above, now for arrayvalued fields
// into the variable 'value'
template <typename T>
struct rfaaccessor_t {
    rfaaccessor_t(const casa::RecordInterface& r, const RecordFieldId& fi) {
        r.get(fi, value);
    }

    casa::Array<T>   value;
};

// function template for converting casa (scalar) valued RecordFields into
// a PyObject
template <typename T, typename U, PyObject* (*fptr)(U)>
struct gen {
    static PyObject* do_gen(const casa::RecordInterface& r, Int i) {
        rfaccessor_t<T>  rfa(r, i);
        return fptr(U(rfa.value));
    }
};

// id. for array-valued thingies
// Specify a 'Transform' template template parameter;
// this class will be used to transform individual array
// elements to a corresponding python-numarray-type
template <typename T, template <typename> class Transform = warp::identity>
struct gena {
    static PyObject* do_gen(const casa::RecordInterface& r, Int i) {
        rfaaccessor_t<T>  rfa(r, i);
        return arraytransformer<Transform>()(rfa.value);
    }
};

// function pointer which takes a "RecordInterface" and a fieldnumber
// and warps it into a PyObject*
typedef PyObject* (*genfunc)(const casa::RecordInterface&, Int num);
typedef map<casa::DataType,genfunc> genmap_t;


PyObject* make_pystr( const casa::String& s ) {
    return PyString_FromString(s.chars());
}

// because of 'circular' dependency, or, recursiveness, if you will,
// the function which turns a TableRecord into a PyDict must be able
// to find itself in the 'genmap_t' typed variable, but from the
// function we must be able to reference the genmap_t typed variable,
// decide to have a placeholder here which we can stick into the map.
PyObject* make_pydict_fromsubrecord(casa::RecordInterface const&, Int num);

PyObject* make_pyobject_fromsubtable(casa::RecordInterface const&, Int num);


// fill in the mapping
genmap_t make_genmap()
{
    genmap_t   rv;

    rv[TpBool] = &gen<casa::Bool,long,PyBool_FromLong>::do_gen;
    rv[TpInt] =  &gen<Int,long,PyInt_FromLong>::do_gen;
    rv[TpUInt] = &gen<uInt,long,PyInt_FromLong>::do_gen;
    rv[TpFloat] = &gen<Float,double,PyFloat_FromDouble>::do_gen;
    rv[TpDouble] = &gen<Double,double,PyFloat_FromDouble>::do_gen;
    rv[TpString] = &gen<casa::String,const casa::String&,make_pystr>::do_gen;

    rv[TpRecord] = &make_pydict_fromsubrecord;
    rv[TpTable]  = &make_pyobject_fromsubtable;

    rv[TpArrayBool] = &gena<casa::Bool>::do_gen;
    rv[TpArrayInt]  = &gena<Int>::do_gen;
    rv[TpArrayFloat] = &gena<Float>::do_gen;
    rv[TpArrayDouble]= &gena<Double>::do_gen;
    rv[TpArrayComplex] = &gena<Complex, warp::ToPyComplex>::do_gen;
    rv[TpArrayDComplex] = &gena<DComplex, warp::ToPyDComplex>::do_gen;
    rv[TpArrayString]   = &gena<String, warp::ToPyStr>::do_gen;

    return rv;
}
genmap_t   genmap = make_genmap();



namespace boost { namespace python {
    // convert  casa String to PyString
    template <>
    struct to_python_value<casa::String const&> :
        detail::builtin_to_python
    {
        inline PyObject* operator()(casa::String const& cs) const
        {
            return ::PyString_FromString(cs.chars());
        }
    };

    // convert casa::Record to PyDict
    template <>
    struct to_python_value<casa::RecordInterface const&>:
        detail::builtin_to_python
    {
        inline PyObject* operator()( casa::RecordInterface const& ri) const
        {
            PyDictObject*   dict = (PyDictObject*)PyDict_New();

            if( !dict ) {
                return NULL;
            }

            // loop over all recordfields and transform them into
            // appropriate PythonObject* 's
            for( uInt fld=0; fld<ri.nfields(); fld++ )
            {
                genmap_t::iterator curgen( genmap.find(ri.type((Int)fld)) );

                if( curgen!=genmap.end() ) {
                    PyDict_SetItem(  (PyObject*)dict,
                                     warp::ToPyStr<>()( ri.name((Int)fld) ),
                                     curgen->second( ri, (Int)fld ) );
                } else {
                    cout << "Unknown fieldtype for " << ri.name((Int)fld)
                         << " (type=" << ri.type((Int)fld) << ")" << endl;
                }
            }
            return (PyObject*)dict;
        }
    };

}} // boost::python



// And do it the other way around ->
// Convert from python string to
// casa::String
struct CasaStringFromPythonString {
    CasaStringFromPythonString()
    {
        converter::registry::push_back(&convertible, &construct, type_id<String>() );
    }
    static void* convertible( PyObject* ptr )
    {
        if( PyString_Check(ptr)==0 )
        {
            return 0;
        }
        return ptr;
    }
    static void construct( ::PyObject* pyptr,
            boost::python::converter::rvalue_from_python_stage1_data* data )
    {
        const char* value = ::PyString_AsString(pyptr);

        if( value==0 )
        {
            boost::python::throw_error_already_set();
        }
        void* storage =
            ((boost::python::converter::rvalue_from_python_storage<casa::String>*)data)->storage.bytes;
        new (storage) casa::String(value);
        data->convertible = storage;
    }
};


//
// access a value in row 'r' in a scalar column 'c'
// in table 't'.
// Put the result into 'value'
// 
template <typename T>
struct scolaccessor {
    scolaccessor(const Table& t, const std::string& c, uInt r)
    {
        ROScalarColumn<T>(t, c).get(r, value);
    }

    T            value;
};

// id. but now for an ArrayColumn
template <typename T>
struct acolaccessor {
    acolaccessor(const Table& t, const std::string& c, uInt r)
    {
        ROArrayColumn<T>(t, c).get(r, value);
    }

    Array<T>       value;
};

// Take all the values in column 'c' in table 't'
// and put the result in 'values'
template <typename T,
          template <typename> class colType = ROScalarColumn,
          template <typename> class arrType = Vector >
struct colaccessor2 {
    colaccessor2(const Table& t, const std::string& c)
    {
        colType<T>(t, c).getColumn(values, True);
    }
    arrType<T>   values;
};


//
// proxy for a column-object
//   Easier to keep state 
// 
struct ColumnProxy {

    ColumnProxy( const Table& t, const std::string& col ):
        tab(t),
        colName(col)
    {
        if( !tab.tableDesc().isColumn(colName) )
        {
            throw column_proxy_no_such_column();
        }
        cd = t.tableDesc().columnDesc(colName);
    }

    unsigned int length( void ) const {
        return tab.nrow();
    }

    PyObject* getCell( int idx )
    {
        // If idx out of range, return None
        // but first see if we must adjust the idx
        if( idx<0 ) {
            idx += (int)tab.nrow();
        }
        // if after adjustment, idx out of range
        // throw up
        if( idx<0 || idx>=tab.nrow() ) {
            PyErr_SetString(PyExc_StopIteration, "");
            return NULL;
        }

        switch( cd.trueDataType() )
        {
            case TpBool:
                {
                    scolaccessor<casa::Bool>    b(tab, colName, idx);
                    return PyBool_FromLong( (long)(b.value==True) );
                }
                break;
            case TpArrayBool:
                {
                    const acolaccessor<casa::Bool>  d(tab, colName, idx);
                    return arraytransformer<>()( d.value );
                }
                break;
            case TpInt:
                {
                    scolaccessor<Int>    b(tab, colName, idx);
                    return PyInt_FromLong( (long)b.value );
                }
                break;
            case TpArrayInt:
                {
                    const acolaccessor<Int>  d(tab, colName, idx);
                    return arraytransformer<>()( d.value );
                }
                break;

            case TpString:
                {
                    scolaccessor<String>  s(tab, colName, idx);
                    return PyString_FromString(s.value.chars());
                }
                break;
            case TpArrayString:
                {
                    const acolaccessor<String>  d(tab, colName, idx);
                    return arraytransformer<warp::ToPyStr>()( d.value );
                }
                break;

            case TpFloat:
                {
                    scolaccessor<Float>  s(tab, colName, idx);
                    return PyFloat_FromDouble((double)s.value);
                }
                break;
            case TpArrayFloat:
                {
                    const acolaccessor<Float>  d(tab, colName, idx);
                    return arraytransformer<>()( d.value );
                }
                break;

            case TpDouble:
                {
                    scolaccessor<Double>  s(tab, colName, idx);
                    return PyFloat_FromDouble(s.value);
                }
                break;
            case TpArrayDouble:
                {
                    const acolaccessor<Double>  d(tab, colName, idx);
                    return arraytransformer<>()( d.value );
                }
                break;

            case TpComplex:
                {
                    scolaccessor<Complex>  s(tab, colName, idx);
                    return PyComplex_FromDoubles(s.value.real(), s.value.imag());
                }
                break;
            case TpArrayComplex:
                {
                    const acolaccessor<Complex>  d(tab, colName, idx);
                    return arraytransformer<warp::ToPyComplex>()( d.value );
                }
                break;

            case TpDComplex:
                {
                    scolaccessor<DComplex>  s(tab, colName, idx);
                    return PyComplex_FromDoubles(s.value.real(), s.value.imag());
                }
                break;
            case TpArrayDComplex:
                {
                    const acolaccessor<DComplex>  d(tab, colName, idx);
                    return arraytransformer<warp::ToPyDComplex>()( d.value );
                }
                break;

            default:
                break;
        }
        // Unsupported type
        throw column_proxy_unsupported_column_datatype();
    }

    // get all the values in the column
    PyObject* get( void )
    {
        switch( cd.trueDataType() )
        {
            case TpBool:
                {
                    const colaccessor2<casa::Bool>    b(tab, colName);
                    return arraytransformer<>()( b.values );
                }
                break;

            case TpInt:
                {
                    const colaccessor2<Int>    b(tab, colName);
                    return arraytransformer<>()( b.values );
                }
                break;
            case TpArrayInt:
                {
                    const colaccessor2<Int,ROArrayColumn,Array>   b(tab, colName);
                    return arraytransformer<>()( b.values );
                }
                break;

            case TpFloat:
                {
                    const colaccessor2<Float>    b(tab, colName);
                    return arraytransformer<>()( b.values );
                }
                break;
            case TpArrayFloat:
                {
                    const colaccessor2<Float,ROArrayColumn, Array>    b(tab, colName);
                    return arraytransformer<>()( b.values );
                }
                break;

            case TpDouble:
                {
                    const colaccessor2<Double>    b(tab, colName);
                    return arraytransformer<>()( b.values );
                }
                break;
            case TpArrayDouble:
                {
                    const colaccessor2<Double,ROArrayColumn, Array>    b(tab, colName);
                    return arraytransformer<>()( b.values );
                }
                break;

            case TpString:
                {
                    colaccessor2<String>     s(tab, colName);
                    return arraytransformer<warp::ToPyStr>()( s.values );
                }
                break;

            case TpArrayComplex:
                {
                    const colaccessor2<Complex, ROArrayColumn, Array>  d(tab, colName);
                    return arraytransformer<warp::ToPyComplex>()( d.values );
                }
                break;
            case TpArrayDComplex:
                {
                    const colaccessor2<DComplex, ROArrayColumn, Array>  d(tab, colName);
                    return arraytransformer<warp::ToPyDComplex>()( d.values );
                }
                break;
        }
        // Unsupported type
        throw column_proxy_unsupported_column_datatype();
    }


    // get the column keywords
    PyObject* keywordSet( void ) 
    {
        return boost::python::to_python_value<RecordInterface const&>()( cd.keywordSet() );
    }

    Table       tab;
    ColumnDesc  cd;
    std::string colName;

    private:
        ColumnProxy();
};

std::ostream& operator<<( std::ostream& os, const ColumnProxy& cp )
{
    return os << cp.cd;
}


//
// Proxy for a casa Table
//    very thin wrapper
//    but makes it easy for us
//    and Boost.Python to bind
//    only specific members and/or
//    do some processing before/after
//    the actual call to the implementation
struct TableProxy {

    // default c'tor
    TableProxy():
        tab(),
        ra( tab )
    {
    }

    // do implement the copy c'tor ->
    // this'un gets called by the BOOST stuff
    // and we must take precautions to 
    // make sure the (refcounted) table
    // and ROTableRow are properly initialized...
    // 
    // It certainly isn't valid to "just copy
    // other.tab and other.ra across" even though
    // both objects support copy construction.
    //
    // The problem is with the ROTableRow, it is,
    // when it is constructed, associated with a
    // specific table. The ROTableRow in a TableProxy
    // should always be 'associated' with the table
    // object in the same object.
    //
    // In the example below, without a "proper" (read: correct)
    // copy c'tor, after executing these statements,
    // TableProxy b's ROTableRow member references
    // TableProxy a's Table and *not* b's Table,
    // which messes up reference counting [which happens
    // inside the Tables] and hence cleaning up of
    // tempfiles etc.
    //
    // TableProxy    a;
    // TableProxy    b(a);
    //
    // A 'correct' copy constructor is implemented below,
    // we can just copy construct our table with other's table
    // but we must explicitly associate our ROTableRow with
    // *our* table-object
    TableProxy( const TableProxy& other ):
        tab( other.tab ),
        ra( tab )
    {
    }
    const TableProxy& operator=( const TableProxy& other )
    {
        if( this!=&other )
        {
            tab = other.tab;
            ra  = ROTableRow(tab);
        }
        return *this;
    }

    TableProxy( const Table& t ):
        tab( t ),
        ra( tab )
    {
    }

    TableProxy( const std::string& s ):
        tab( s ),
        ra( tab )
    {
    }
    
    unsigned int ncol( void ) const
    {
        return tab.tableDesc().ncolumn();
    }

    unsigned int nrow( void ) const {
        return tab.nrow();
    }

    PyObject* colnames( void ) const {
        Vector<String>  cn( tab.tableDesc().columnNames() );
        return arraytransformer<warp::ToPyStr>()((const Array<String>&)cn);
    }
    
    PyObject* getRow( int r ) {
        static  boost::python::to_python_value<RecordInterface const&>  warper;
        if( r<0 ) {
            r += (int)tab.nrow();
        }
        if( r<0 || r>=(int)tab.nrow() ) {
            PyErr_SetString(PyExc_StopIteration, "");
            return NULL;
        }
        return warper(ra.get(r));
    }

    ColumnProxy* getColumn( const std::string& colnm )
    {
        return new ColumnProxy(tab, colnm);
    }

    const casa::String& tableName( void ) const {
        return tab.tableName();
    }

    PyObject* keywordSet( void ) 
    {
        return boost::python::to_python_value<RecordInterface const&>()( tab.keywordSet() );
    }

    TableProxy  query( const std::string& taql ) {
        return TableProxy( tableCommand(String("select from $1 where ")+String(taql), tab) );

    }

    // query with select clause
    TableProxy query( const std::string& cols,
                      const std::string& where )
    {
        String  qry("select ");

        if( cols.size() )
            qry += String(cols);
        qry += " from $1 ";
        if( where.size() )
            qry += String("where ")+String(where);
        return TableProxy( tableCommand(qry, tab) );
    }

    Table       tab;
    ROTableRow  ra;
};

// Create function pointers to both versions of TableProxy::query
// in order to help Boost decide which to select...
TableProxy (TableProxy::*q1)(const std::string&) = &TableProxy::query;
TableProxy (TableProxy::*q2)(const std::string&, const std::string&) = &TableProxy::query;

//
// The casa module
// 
BOOST_PYTHON_MODULE(pycasa)
{
    // define a table
    class_<TableProxy>("table")
        .def(init<>())
        .def(init<std::string>())
        .def("tableName", &TableProxy::tableName,
                          return_value_policy<copy_const_reference>())
        .def("__len__", &TableProxy::nrow)
        .def("ncol", &TableProxy::ncol)
        .def("getcol", &TableProxy::getColumn,
                       return_internal_reference<>())
        .def("__getitem__", &TableProxy::getRow)
        .def("query", q1)
        .def("query", q2)
        .def("colnames", &TableProxy::colnames)
        .def("keywordSet", &TableProxy::keywordSet)
        ;
        
    class_<ColumnProxy>("column", no_init)
        .def("__str__", &stringifier<ColumnProxy>::asString )
        .def("__getitem__", &ColumnProxy::getCell )
        .def("get", &ColumnProxy::get)
        .def("__len__",&ColumnProxy::length)
        .def("keywordSet", &ColumnProxy::keywordSet)
        ;

    //
    // Do register type conversions
    //
    CasaStringFromPythonString();
}

//
// Transform a recordinterface based field into a PyDict
// Need this wrapper since it must be defined *after*
// the boost::python::to_python_value<RecordInterface> is defined
//
PyObject* make_pydict_fromsubrecord(casa::RecordInterface const& r, Int num) {
    // retrieve pointer to the subrecord
    const casa::RecordInterface& srec = *((RecordInterface*)(r.get_pointer(num, TpRecord)));

    // and warp it into a PyObject
    return boost::python::to_python_value<casa::RecordInterface const&>()(srec);
}

//
// Define this only after the class_<TableProxy> is defined 
// such that we can actually give that back
//
PyObject* make_pyobject_fromsubtable(casa::RecordInterface const& r, Int num) {
    return PyString_FromString("casa::Table");
#if 0
    // retrieve pointer to the subtable
    const casa::TableKeyword& trec = *((const TableKeyword*)(r.get_pointer(num, TpTable)));
    // Prepare return value, it's going to be one of our own defined
    // python classes (the tableproxy thingy) and we get a reference to the
    // held C++ object of type 'TableProxy'. Then we set the value to the
    // new table and return the python object
    object      rv = ( class_<TableProxy>("table").def(init<>()) )();
    TableProxy& tbref = extract<TableProxy&>(rv);

    tbref = trec.table();
    // and warp it into a PyObject
    incref(rv.ptr());
    return rv.ptr();
#endif
}
